import sys
import typing as tp
from copy import copy
from dataclasses import dataclass, field
from enum import Enum
from functools import total_ordering

import jax
import jax.numpy as jnp
import jax.tree_util
import numpy as np
import treex as tx

EPSILON = 1e-7
F = tp.TypeVar("F", bound=tp.Callable)


KeySeq = tx.KeySeq

T = tp.TypeVar("T")
Container = tp.Union[
    T,
    tp.Tuple["Container", ...],
    tp.Dict[str, "Container"],
]
ArrayHolder = tp.Union[Container[np.ndarray], np.ndarray]

IndexLike = tp.Union[str, int, tp.Iterable[tp.Union[str, int]]]

Shape = tp.Sequence[int]
ShapeLike = tp.Union[int, Shape]
FloatLike = tp.Union[float, np.ndarray]
DType = tp.Any
ParamName = str

Params = tp.Mapping[str, tp.Mapping[ParamName, np.ndarray]]
State = tp.Mapping[str, tp.Mapping[str, np.ndarray]]
PadFn = tp.Callable[[int], tp.Tuple[int, int]]
PadFnOrFns = tp.Union[PadFn, tp.Sequence[PadFn]]
PRNGKey = np.ndarray
Parameters = tp.Dict[str, tp.Any]
Labels = tp.Mapping[str, tp.Any]
ParameterCollection = tp.Dict[str, Parameters]
Logs = tp.Dict[str, jnp.ndarray]
Outputs = tp.Any
Inputs = tp.Any
Index = tp.Union[int, str]
Path = tp.Tuple[Index, ...]
Grads = tp.Any
RNG = tp.Union[KeySeq, np.ndarray]
Scalar = tp.Union[np.ndarray, float, int]
SummaryModule = tp.Any
SummaryValue = tp.Any
KeyLike = tp.Union[int, jnp.ndarray]

NetParams = tp.Any
NetStates = tp.Any
ModuleParams = tp.Any
ModuleStates = tp.Any
MetricsStates = tp.Any
OptimizerStates = tp.Any
OptimizerStates = tp.Any
Grads = tp.Any
Pytree = tp.Any


class MissingModule(Exception):
    pass


class MissingOptimizer(Exception):
    pass


class MissingMethod(Exception):
    pass


class DependencyUnavailable(Exception):
    pass


class ShapeMismatch(Exception):
    pass


class MissingParameter(Exception):
    pass


class NoContext(Exception):
    pass


class ModuleOrderError(Exception):
    pass


class SubmoduleNotRegistered(Exception):
    pass


class ModelNotInitialized(Exception):
    pass



from .data_handler import DataHandler
from .dataset import DataLoader, Dataset
from .utils import (
    map_append,
    map_structure,
    train_validation_split,
    unpack_x_y_sample_weight,
)

__all__ = ["Dataset", "DataLoader"]



# Implementation based on tf.keras.callbacks.py
# https://github.com/tensorflow/tensorflow/blob/v2.2.0/tensorflow/python/keras/callbacks.py
import enum
import logging
import signal
import typing as tp

import numpy as np

from .callback import Callback

ORIGINAL_HANDLER = signal.getsignal(signal.SIGINT)


class SigIntMode(enum.Enum):
    TRAIN = enum.auto()
    TEST = enum.auto()
    PREDICT = enum.auto()


class SigInt(Callback):
    def __init__(self, mode: tp.Union[SigIntMode, str]):
        super().__init__()
        self.mode = mode if isinstance(mode, SigIntMode) else SigIntMode(mode.upper())

    def signal_handler(self, signal, frame):
        print("\n\nStopping...\n\n")
        self.model.stop_training = True
        # signal.signal(signal.SIGINT, ORIGINAL_HANDLER)

    def on_train_begin(self, logs=None):
        signal.signal(signal.SIGINT, self.signal_handler)

    def on_train_end(self, logs=None):
        if self.mode == SigIntMode.TRAIN:
            signal.signal(signal.SIGINT, ORIGINAL_HANDLER)

    def on_test_begin(self, logs=None):
        signal.signal(signal.SIGINT, self.signal_handler)

    def on_test_end(self, logs=None):
        if self.mode == SigIntMode.TEST:
            signal.signal(signal.SIGINT, ORIGINAL_HANDLER)

    def on_predict_begin(self, logs=None):
        signal.signal(signal.SIGINT, self.signal_handler)

    def on_predict_end(self, logs=None):
        if self.mode == SigIntMode.PREDICT:
            signal.signal(signal.SIGINT, ORIGINAL_HANDLER)



import typing as tp

import flax.linen as nn
import jax
import jax.numpy as jnp
import treeo as to
from attr import mutable
from flax.core.frozen_dict import FrozenDict

from elegy import types, utils

FrozerVariables = FrozenDict[str, tp.Mapping[str, tp.Any]]
Variables = tp.Mapping[str, tp.Mapping[str, tp.Any]]
M = tp.TypeVar("M", bound="nn.module.Module")
RNGDict = tp.Dict[str, jnp.ndarray]


class ModuleState(tp.Generic[M], to.Tree, to.Immutable):
    variables: tp.Optional[FrozerVariables] = to.node()

    hashable_module: to.Hashable[M] = to.static()
    mutable_train: tp.Tuple[str] = to.static()
    mutable_eval: tp.Tuple[str] = to.static()
    rngs_init: tp.Tuple[str] = to.static()
    rngs_apply: tp.Tuple[str] = to.static()
    method_init: str = to.static()

    def __init__(
        self,
        module: M,
        variables: tp.Optional[Variables] = None,
        mutable_train: tp.Sequence[str] = ("batch_stats", "cache"),
        mutable_eval: tp.Optional[tp.Sequence[str]] = None,
        rngs_init: tp.Sequence[str] = ("params",),
        rngs_apply: tp.Sequence[str] = ("dropout",),
        method_init: str = "__call__",
    ):

        self.hashable_module = to.Hashable(module)
        self.variables = FrozenDict(variables) if variables is not None else None
        self.mutable_train = tuple(mutable_train)
        self.mutable_eval = (
            tuple(mutable_eval) if mutable_eval is not None else tuple(mutable_train)
        )
        self.rngs_init = tuple(rngs_init)
        self.rngs_apply = tuple(rngs_apply)
        self.method_init = method_init

    @property
    def module(self: "ModuleState[M]") -> M:
        return self.hashable_module.value

    @property
    def initialized(self: "ModuleState[M]") -> bool:
        return self.variables is not None

    def init(
        self: "ModuleState[M]",
        key: tp.Optional[types.KeyLike],
        *args,
        training: bool = True,
        rngs: tp.Optional[tp.Union[RNGDict, tp.Sequence[str]]] = None,
        **kwargs,
    ) -> "ModuleState[M]":
        """
        Initialize the module.
        """

        module_manager: ModuleState[M] = self

        if "method" not in kwargs:
            method = getattr(module_manager.module, self.method_init)
        else:
            method = kwargs.pop("method")

        arg_names = utils._function_argument_names(method)

        if arg_names is not None and "training" in arg_names:
            kwargs["training"] = training

        # calculate rngs
        if isinstance(rngs, dict):
            pass
        else:
            if key is not None:
                key = utils.Key(key)

                if rngs is None:
                    rng_collections = self.rngs_init
                else:
                    rng_collections = rngs

                rngs = _split_into_collection(key, rng_collections)
            else:
                rngs = {}

        variables = module_manager.module.init(
            rngs,
            *args,
            method=method,
            **kwargs,
        )

        if not isinstance(variables, FrozenDict):
            variables = FrozenDict(variables)

        module_manager = module_manager.replace(
            variables=variables,
            hashable_module=to.Hashable(module_manager.module),
        )

        return module_manager

    def apply(
        self: "ModuleState[M]",
        key: tp.Optional[types.KeyLike],
        *args,
        training: bool = True,
        rngs: tp.Optional[tp.Union[RNGDict, tp.Sequence[str]]] = None,
        mutable: tp.Optional[tp.Union[bool, tp.Sequence[str]]] = None,
        **kwargs,
    ) -> tp.Tuple[tp.Any, "ModuleState[M]"]:
        return self._forward(
            self.module.__call__,
            key,
            *args,
            training=training,
            rngs=rngs,
            mutable=mutable,
            **kwargs,
        )

    def stateless(
        self: "ModuleState[M]",
        key: tp.Optional[types.KeyLike],
        *args,
        training: bool = True,
        rngs: tp.Optional[tp.Union[RNGDict, tp.Sequence[str]]] = None,
        mutable: tp.Optional[tp.Union[bool, tp.Sequence[str]]] = None,
        **kwargs,
    ) -> tp.Any:
        return self.apply(
            key,
            *args,
            training=training,
            rngs=rngs,
            mutable=mutable,
            **kwargs,
        )[0]

    def _forward(
        self: "ModuleState[M]",
        method: tp.Callable,
        key: tp.Optional[types.KeyLike],
        *args,
        training: bool = True,
        rngs: tp.Optional[tp.Union[RNGDict, tp.Sequence[str]]] = None,
        mutable: tp.Optional[tp.Union[bool, tp.Sequence[str]]] = None,
        **kwargs,
    ) -> tp.Tuple[tp.Any, "ModuleState[M]"]:

        variables = self.variables

        if variables is None:
            raise ValueError(f"'variables' field is not set for module: {self.module}")

        arg_names = utils._function_argument_names(method)

        if arg_names is not None and "training" in arg_names:
            kwargs["training"] = training

        # calculate rngs
        if isinstance(rngs, dict):
            pass
        else:
            if key is not None:
                key = utils.Key(key)

                if rngs is None:
                    rng_collections = self.rngs_apply
                else:
                    rng_collections = rngs

                rngs = _split_into_collection(key, rng_collections)
            else:
                rngs = {}

        # calculate mutable
        if mutable is None:
            mutable = self.mutable_train if training else self.mutable_eval

        output, updates = self.module.apply(
            variables,
            *args,
            rngs=rngs,
            method=method,
            mutable=mutable,
            **kwargs,
        )

        self = self.replace(
            variables=variables.copy(updates),
        )

        return output, self

    def __getitem__(self, key: str) -> tp.Any:
        if self.variables is None:
            raise KeyError(f"'variables' field is not set for module: {self.module}")

        return self.variables[key]

    def __contains__(self, key: str) -> bool:
        if self.variables is None:
            raise KeyError(f"'variables' field is not set for module: {self.module}")

        return key in self.variables

    def update(self: "ModuleState[M]", **kwargs) -> "ModuleState[M]":
        if self.variables is None:
            raise ValueError(f"'variables' field is not set for module: {self.module}")

        return self.replace(variables=self.variables.copy(kwargs))


def _split_into_collection(
    key: jnp.ndarray,
    collection: tp.Sequence[str],
) -> tp.Dict[str, jnp.ndarray]:
    """
    Split the key into the specified rngs.
    """

    rngs = jax.random.split(key, len(collection))

    keys_collection = {col: rngs[i] for i, col in enumerate(collection)}

    return keys_collection



import enum
import pathlib
import typing as tp
from abc import ABC, abstractmethod
from dataclasses import dataclass

import cloudpickle
import einops
import jax
import jax.numpy as jnp
import numpy as np
import treex as tx

from elegy import types, utils

from . import utils as model_utils

try:
    import tensorflow as tf
    from jax.experimental import jax2tf
except ImportError:
    tf = None
    jax2tf = None

A = tp.TypeVar("A")
M = tp.TypeVar("M", bound="ModelCore")
Me = tp.TypeVar("Me", bound="tx.Metric")

PredStepOutput = tp.Tuple[tp.Any, M]
TestStepOutput = tp.Tuple[jnp.ndarray, types.Logs, M]
LossStepOutput = tp.Tuple[jnp.ndarray, tp.Tuple[types.Logs, M]]
GradStepOutput = tp.Tuple[M, types.Logs, M]
TrainStepOutput = tp.Tuple[types.Logs, M]

InitStep = tp.Callable[[M, jnp.ndarray, tp.Any], M]
PredStep = tp.Callable[[M, tp.Any], PredStepOutput[M]]
TestStep = tp.Callable[[M, tp.Any, types.Labels], TestStepOutput[M]]
GradStep = tp.Callable[[M, tp.Any, types.Labels], GradStepOutput[M]]
TrainStep = tp.Callable[[M, tp.Any, types.Labels], TrainStepOutput[M]]


# --------------------------------------------------------------------------------
# DistributedStrategy
# --------------------------------------------------------------------------------


class DistributedStrategy(ABC):
    def from_local(self, model: M) -> M:
        return model

    def to_local(self, model: M) -> M:
        return model

    def lift_data(self, data: A) -> A:
        return data

    def lift_key(self, key: jnp.ndarray) -> jnp.ndarray:
        return key

    def lift_batch_size(self, batch_size: int) -> int:
        return batch_size

    def handle_post_init(self, model: M) -> M:
        return model

    def handle_metrics(
        self,
        metrics: Me,
    ) -> Me:
        return metrics

    def handle_model_and_grads(
        self,
        model: M,
        grads: M,
    ) -> tp.Tuple[M, M]:
        return model, grads

    @abstractmethod
    def init_step_fn(self, model: M) -> InitStep[M]:
        ...

    @abstractmethod
    def pred_step_fn(self, model: M) -> PredStep[M]:
        ...

    @abstractmethod
    def test_step_fn(self, model: M) -> TestStep[M]:
        ...

    @abstractmethod
    def train_step_fn(self, model: M) -> TrainStep[M]:
        ...

    # implement order methods, required so that DistributedStrategy can be
    # used as a key in a dict
    def __lt__(self, other):
        return self.__class__.__name__ < other.__class__.__name__

    def __le__(self, other):
        return self.__class__.__name__ <= other.__class__.__name__

    def __gt__(self, other):
        return self.__class__.__name__ > other.__class__.__name__

    def __ge__(self, other):
        return self.__class__.__name__ >= other.__class__.__name__


@dataclass(unsafe_hash=True)
class Eager(DistributedStrategy):
    def init_step_fn(self, model: M) -> InitStep[M]:
        return model.__class__._static_init_step

    def pred_step_fn(self, model: M) -> PredStep[M]:
        return model.__class__._static_pred_step

    def test_step_fn(self, model: M) -> TestStep[M]:
        return model.__class__._static_test_step

    def train_step_fn(self, model: M) -> TrainStep[M]:
        return model.__class__._static_train_step


@dataclass(unsafe_hash=True)
class JIT(DistributedStrategy):
    # donate 'model' memory buffer since we return an updated model
    def init_step_fn(self, model: M) -> InitStep[M]:
        return jax.jit(model.__class__._static_init_step, donate_argnums=0)

    def pred_step_fn(self, model: M) -> PredStep[M]:
        return jax.jit(model.__class__._static_pred_step, donate_argnums=0)

    def test_step_fn(self, model: M) -> TestStep[M]:
        return jax.jit(model.__class__._static_test_step, donate_argnums=0)

    def train_step_fn(self, model: M) -> TrainStep[M]:
        return jax.jit(model.__class__._static_train_step, donate_argnums=0)


@dataclass(unsafe_hash=True)
class DataParallel(DistributedStrategy):
    def from_local(self, model: M) -> M:
        # device_idxs used to inform pmap about the number of devices
        device_idxs = jnp.arange(jax.device_count())
        model = jax.pmap(
            lambda idx, model: model,
            in_axes=(0, None),
            out_axes=0,
        )(device_idxs, model)

        return model

    def to_local(self, model: M) -> M:
        return jax.tree_map(lambda x: x[0], model)

    def lift_data(self, data: A) -> A:
        data = jax.tree_map(
            lambda x: einops.rearrange(
                x,
                "(device batch) ... -> device batch ...",
                device=jax.device_count(),
            ),
            data,
        )
        return data

    def lift_key(self, key: jnp.ndarray) -> jnp.ndarray:
        key = einops.repeat(
            key,
            "... -> device ...",
            device=jax.device_count(),
        )
        return key

    def lift_batch_size(self, batch_size: int) -> int:
        return batch_size * jax.device_count()

    def handle_post_init(self, model: M) -> M:
        return model.map(
            lambda key: jax.random.fold_in(key, jax.lax.axis_index("device")),
            tx.Rng,
            inplace=True,
        )

    def handle_metrics(
        self,
        metrics: Me,
    ) -> Me:
        metrics = jax.lax.stop_gradient(metrics)
        metrics = jax.lax.psum(metrics, axis_name="device")
        return metrics

    def handle_model_and_grads(
        self,
        model: M,
        grads: M,
    ) -> tp.Tuple[M, M]:
        grads = jax.lax.pmean(grads, axis_name="device")
        model = model.map(lambda x: jax.lax.pmean(x, axis_name="device"), tx.BatchStat)

        return model, grads

    def init_step_fn(self, model: M) -> InitStep[M]:
        return jax.pmap(
            model.__class__._static_init_step,
            axis_name="device",
            donate_argnums=0,
        )

    def pred_step_fn(self, model: M) -> PredStep[M]:
        return jax.pmap(
            model.__class__._static_pred_step,
            axis_name="device",
            donate_argnums=0,
        )

    def test_step_fn(self, model: M) -> TestStep[M]:
        return jax.pmap(
            model.__class__._static_test_step,
            axis_name="device",
            out_axes=(0, None, 0),  # None = logs not replicated
            donate_argnums=0,
        )

    def train_step_fn(self, model: M) -> TrainStep[M]:
        return jax.pmap(
            model.__class__._static_train_step,
            axis_name="device",
            out_axes=(None, 0),  # None = logs not replicated
            donate_argnums=0,
        )


# --------------------------------------------------------------------------------
# ModelCore
# --------------------------------------------------------------------------------


class ModelMeta(tx.TreeMeta):
    def __call__(self, *args, **kwargs) -> "ModelCore":
        model: ModelCore = super().__call__(*args, **kwargs)

        model.create_jit_functions()

        return model


class ModelCore(tx.Treex, tx.Filters, metaclass=ModelMeta):

    seed: tp.Union[int, jnp.ndarray] = 42
    _initialized: bool = False
    _distributed_strategy: DistributedStrategy = JIT()

    init_step_fn: tp.Dict[DistributedStrategy, InitStep]
    pred_step_fn: tp.Dict[DistributedStrategy, PredStep]
    test_step_fn: tp.Dict[DistributedStrategy, TestStep]
    train_step_fn: tp.Dict[DistributedStrategy, TrainStep]

    def __init__(
        self,
        eager: bool = False,
        seed: int = 42,
    ):
        self.seed = seed

        self._distributed_strategy = Eager() if eager else JIT()

    @property
    def initialized(self) -> bool:
        return self._initialized

    @property
    def is_distributed(self) -> bool:
        return (
            self._distributed_strategy != JIT()
            and self._distributed_strategy != Eager()
        )

    @property
    def is_eager(self) -> bool:
        return self._distributed_strategy == Eager()

    @property
    def distributed_strategy(self) -> DistributedStrategy:
        return self._distributed_strategy

    def create_jit_functions(self):
        with tx.make_mutable(self, toplevel_only=True):
            cls = self.__class__
            self._jitted_members: tp.Set[str] = set()

            self.init_step_fn = {}
            self.pred_step_fn = {}
            self.test_step_fn = {}
            self.train_step_fn = {}

            self._maybe_build_strategy()

            self._jitted_members |= {
                "init_step_fn",
                "pred_step_fn",
                "test_step_fn",
                "train_step_fn",
            }

    def _maybe_build_strategy(self):
        strategy = self._distributed_strategy

        if (
            strategy not in self.init_step_fn
            or strategy not in self.pred_step_fn
            or strategy not in self.test_step_fn
            or strategy not in self.train_step_fn
        ):
            # build strategy functions
            self.init_step_fn[strategy] = strategy.init_step_fn(self)
            self.pred_step_fn[strategy] = strategy.pred_step_fn(self)
            self.test_step_fn[strategy] = strategy.test_step_fn(self)
            self.train_step_fn[strategy] = strategy.train_step_fn(self)

    def __setstate__(self, d):
        with tx.make_mutable(self, toplevel_only=True):
            self.__dict__ = d
            self.create_jit_functions()

    def __getstate__(self):
        d = self.__dict__.copy()

        # remove jitted functions
        for member in self._jitted_members:
            if member in d:
                del d[member]

        return d

    def _update_from(self: M, other: M):
        self.__dict__.update(other.__dict__)

    # ----------------------------------------------------------------
    # Abstract API
    # ----------------------------------------------------------------

    def init_step(
        self: M,
        key: jnp.ndarray,
        inputs: tp.Any,
    ) -> M:
        raise types.MissingMethod()

    def pred_step(
        self: M,
        inputs: tp.Any,
    ) -> PredStepOutput[M]:
        raise types.MissingMethod()

    def test_step(
        self: M,
        inputs: tp.Any,
        labels: tp.Mapping[str, tp.Any],
    ) -> TestStepOutput[M]:
        raise types.MissingMethod()

    def grad_step(
        self: M,
        inputs: tp.Any,
        labels: tp.Mapping[str, tp.Any],
    ) -> GradStepOutput[M]:
        raise types.MissingMethod()

    def train_step(
        self: M,
        inputs: tp.Any,
        labels: tp.Mapping[str, tp.Any],
    ) -> TrainStepOutput[M]:
        raise types.MissingMethod()

    def reset_step(self) -> None:
        raise types.MissingMethod()

    # ----------------------------------------------------------------
    # static version
    # ----------------------------------------------------------------

    @staticmethod
    def _static_init_step(
        model: M,
        key: jnp.ndarray,
        inputs: tp.Any,
    ) -> M:
        return model.init_step(key, inputs)

    @staticmethod
    def _static_pred_step(
        model: M,
        inputs: tp.Any,
    ) -> PredStepOutput[M]:
        return model.pred_step(inputs)

    @staticmethod
    def _static_test_step(
        model: M,
        inputs: tp.Any,
        labels: tp.Mapping[str, tp.Any],
    ) -> TestStepOutput[M]:
        return model.test_step(inputs, labels)

    @staticmethod
    def _static_train_step(
        model: M,
        inputs: tp.Any,
        labels: tp.Mapping[str, tp.Any],
    ) -> TrainStepOutput[M]:
        return model.train_step(inputs, labels)

    # ----------------------------------------------------------------
    # distributed API
    # ----------------------------------------------------------------

    def distributed(
        self: M,
        strategy: tp.Optional[DistributedStrategy] = None,
        *,
        inplace: bool = False,
    ) -> M:
        model: M

        if strategy is None:
            strategy = DataParallel()

        if not inplace:
            model = self.copy()
        else:
            model = self

        if strategy == model.distributed_strategy:
            return model

        model = self._distributed_strategy.to_local(model)
        model = strategy.from_local(model)

        model._distributed_strategy = strategy
        model._maybe_build_strategy()

        if inplace:
            self._update_from(model)
            model = self

        return model

    def local(self: M, inplace: bool = False) -> M:
        return self.distributed(JIT(), inplace=inplace)

    def eager(self: M, inplace: bool = False) -> M:
        return self.distributed(Eager(), inplace=inplace)

    # ----------------------------------------------------------------
    # high-level methods
    # ----------------------------------------------------------------

    def init_on_batch(self, inputs: tp.Any):

        key = tx.Key(self.seed)

        inputs = self._distributed_strategy.lift_data(inputs)
        key = self._distributed_strategy.lift_key(key)

        init_step_fn = self.init_step_fn[self._distributed_strategy]
        model = init_step_fn(self, key, inputs)

        if not isinstance(model, type(self)):
            raise ValueError(
                f"Model.init_step() must return an instance of {type(self)}."
            )

        self._update_from(model)
        self._initialized = True

    def predict_on_batch(self, inputs: tp.Any) -> tp.Any:
        """
        Returns predictions for a single batch of samples.

        Arguments:
            x: Input data. A Numpy/Jax array (or array-like), or possibly
                nested python structure of dict, list, tuple that contain
                arrays as leafs.

        Returns:
            Jax array(s) of predictions.

        Raises:
            ValueError: In case of mismatch between given number of inputs and
                expectations of the model.
        """
        if not self._initialized:
            self.init_on_batch(inputs)

        # do this after init_on_batch()
        inputs = self._distributed_strategy.lift_data(inputs)

        self.eval(inplace=True)

        pred_step_fn = self.pred_step_fn[self._distributed_strategy]
        y_pred, model = pred_step_fn(self, inputs)

        if not isinstance(model, type(self)):
            raise ValueError(
                f"Model.pred_step() must return an instance of {type(self)}."
            )

        self._update_from(model)

        return y_pred

    def test_on_batch(
        self,
        inputs: tp.Any,
        labels: tp.Any,
    ) -> types.Logs:
        """
        Test the model on a single batch of samples.

        Arguments:
            x: Input data. It could be:

                - A Numpy array (or array-like), or a list
                    of arrays (in case the model has multiple inputs).
                - A dict mapping input names to the corresponding arrays, if
                    the model has named inputs.
            y: Target data. Like the input data `x`, it could be either Numpy
                array(s) or Jax array(s).
            sample_weight: Optional array of the same length as x, containing
                weights to apply to the model's loss for each sample. In the case of
                temporal data, you can pass a 2D array with shape (samples,
                sequence_length), to apply a different weight to every timestep of
                every sample.

        Returns:
            A `logs` dictionary of containing the main `loss` as well as all
            other losses and metrics.
        Raises:
            ValueError: In case of invalid user-provided arguments.
        """
        if not self._initialized:
            self.init_on_batch(inputs)

        # do this after init_on_batch()
        inputs, labels = self._distributed_strategy.lift_data((inputs, labels))

        if not isinstance(labels, tp.Mapping):
            labels = dict(target=labels)

        self.eval(inplace=True)

        test_step_fn = self.test_step_fn[self._distributed_strategy]
        loss, logs, model = test_step_fn(
            self,
            inputs,
            labels,
        )

        if not isinstance(model, type(self)):
            raise ValueError(
                f"Model.test_step() must return an instance of {type(self)}."
            )

        self._update_from(model)

        return logs

    def train_on_batch(
        self,
        inputs: tp.Any,
        labels: tp.Any,
    ) -> types.Logs:
        """
        Runs a single gradient update on a single batch of data.

        Arguments:
            x: Input data. It could be:

                - A Numpy array (or array-like), or a iterable of arrays
                    (in case the model has multiple inputs).
                - A dict mapping input names to the corresponding arrays,
                    if the model has named inputs.
            y: Target data. Like the input data `x`, it could be either Numpy
                array(s) or Jax array(s). It should be consistent with `x`
                (you cannot have Numpy inputs and array targets, or inversely).
            sample_weight: Optional array of the same length as x, containing
                weights to apply to the model's loss for each sample. In the case of
                temporal data, you can pass a 2D array with shape (samples,
                sequence_length), to apply a different weight to every timestep of
                every sample.
            class_weight: Optional dictionary mapping class indices (integers) to a
                weight (float) to apply to the model's loss for the samples from this
                class during training. This can be useful to tell the model to "pay
                more attention" to samples from an under-represented class.

        Returns:
            A `logs` dictionary of containing the main `loss` as well as all
            other losses and metrics.

        Raises:
            ValueError: In case of invalid user-provided arguments.
        """

        if not self.initialized:
            self.init_on_batch(inputs)

        # do this after init_on_batch()
        inputs, labels = self._distributed_strategy.lift_data((inputs, labels))

        self.train(inplace=True)

        if not isinstance(labels, tp.Mapping):
            labels = dict(target=labels)

        train_step_fn = self.train_step_fn[self._distributed_strategy]
        logs, model = train_step_fn(self, inputs, labels)

        if not isinstance(model, type(self)):
            raise ValueError(
                f"Model.train_step() must return an instance of {type(self)}."
            )

        self._update_from(model)

        return logs

    def save(
        self,
        path: tp.Union[str, pathlib.Path],
    ) -> None:
        """
        Saves the model to disk.

        It creates a directory that includes:

        - `{path}/model.pkl`: The `Model` object instance serialized with `pickle`,
            this allows you to re-instantiate the model later.

        This allows you to save the entirety of the states of a model
        in a directory structure which can be fully restored via
        `Model.load` if the model is already instiated or `elegy.model.load`
        to load the model instance from its pickled version.

        ```python
        import elegy

        model.save('my_model')  # creates folder at 'my_model'
        del model  # deletes the existing model

        # returns a model identical to the previous one
        model = elegy.model.load('my_model')
        ```
        Arguments:
            path: path where model structure will be saved.
        """
        model = self.local()

        if isinstance(path, str):
            path = pathlib.Path(path)

        path.mkdir(parents=True, exist_ok=True)

        with open(path / "model.pkl", "wb") as f:
            cloudpickle.dump(model, f)

    def load(
        self,
        path: tp.Union[str, pathlib.Path],
    ) -> None:
        """
        Loads all weights + states from a folder structure.

        You can load states from other models that have slightly different architecture
        as long as long as it preserves the ordering of the `haiku.Params` + `haiku.State`
        structures, adding or removing layers is fine as long as they don't have weights,
        new layers with weights will be initialized from scratch.

        Arguments:
            path: path to a saved model's directory.
        """
        pass

    def saved_model(
        self,
        inputs: tp.Any,
        path: tp.Union[str, pathlib.Path],
        batch_size: tp.Union[int, tp.Sequence[int], None] = None,
    ):
        """
        Serializes the prediction function of the Model (`pred_step`) as a TensorFlow SavedModel via
        `jax2tf`.

        !!! Note
            Due to a current limitation in JAX it is not possible to create dynamicly
            shaped SavedModels so you must specify the `batch_size` argument to create
            one or more statically shaped versions / signatures: [jax#5915](https://github.com/google/jax/issues/5915).

        Arguments:
            x: A sample input used to infer shapes.
            path: The path where the SavedModel should be saved.
            batch_size: An integer or sequence of integers specifying the size of the batch
                dimension of each of the resulting SavedModel's signatures.

        """
        model = self.local()

        if not model.initialized:
            model.init_on_batch(inputs)

        if model_utils.convert_and_save_model is None:
            raise ImportError(f"Could not import tensorflow.")

        assert jax2tf is not None

        if isinstance(batch_size, int):
            batch_size = [batch_size]

        if isinstance(path, str):
            path = pathlib.Path(path)

        path.mkdir(parents=True, exist_ok=True)

        inputs = jax.tree_map(jnp.asarray, inputs)

        if batch_size is None:
            input_signatures = [
                jax.tree_map(
                    lambda p: tf.TensorSpec(shape=(None,) + p.shape[1:], dtype=p.dtype),
                    inputs,
                )
            ]
            shape_polymorphic_input_spec = jax.tree_map(
                lambda p: jax2tf.shape_poly.PolyShape(
                    "batch", *(["_"] * (len(p.shape) - 1))
                ),
                inputs,
            )
        else:
            input_signatures = [
                jax.tree_map(
                    lambda p: tf.TensorSpec(
                        shape=(batch_size,) + p.shape[1:], dtype=p.dtype
                    ),
                    inputs,
                )
                for batch_size in batch_size
            ]
            shape_polymorphic_input_spec = None

        flat_states, states_def = jax.tree_flatten(model)

        def jax_fn(flat_states, inputs):
            model: ModelCore = jax.tree_unflatten(states_def, flat_states)

            y_pred, _ = model.pred_step(
                inputs=inputs,
            )

            return y_pred

        model_utils.convert_and_save_model(
            jax_fn,
            flat_states,
            str(path),
            input_signatures=input_signatures,
            shape_polymorphic_input_spec=shape_polymorphic_input_spec,
            with_gradient=False,
            enable_xla=True,
            compile_model=False,
            save_model_options=None,
        )



# Implementation based on tf.keras.callbacks.py
# https://github.com/tensorflow/tensorflow/blob/v2.2.0/tensorflow/python/keras/callbacks.py

import typing as tp

import jax.numpy as jnp

import elegy


def default(method):
    """Decorates a method to detect overrides in subclasses."""
    method._is_default = True  # pylint: disable=protected-access
    return method


def is_default(method):
    """Check if a method is decorated with the `default` wrapper."""
    return getattr(method, "_is_default", False)


class Callback(object):
    """Abstract base class used to build new callbacks.

    The `logs` dictionary that callback methods
    take as argument will contain keys for quantities relevant to
    the current batch or epoch.

    Currently, the `.fit()` method of the `Model` class
    will include the following quantities in the `logs` that
    it passes to its callbacks:

    ```python
    on_epoch_end: logs include `acc` and `loss`, and
        optionally include `val_loss`
        (if validation is enabled in `fit`), and `val_acc`
        (if validation and accuracy monitoring are enabled).
    on_train_batch_begin: logs include `size`,
        the number of samples in the current batch.
    on_train_batch_end: logs include `loss`, and optionally `acc`
        (if accuracy monitoring is enabled).
    ```

    Attributes:
        params (dict):  Training parameters
            (eg. verbosity, batch size, number of epochs...).
        model (elegy.model.Model): Reference of the model being trained.
    """

    model: tp.Optional["elegy.model_full.Model"]

    __all__ = [
        "on_epoch_begin",
        "on_epoch_end",
        "on_predict_batch_begin",
        "on_predict_batch_end",
        "on_predict_begin",
        "on_predict_end",
        "on_test_batch_begin",
        "on_test_batch_end",
        "on_test_begin",
        "on_test_end",
        "on_train_batch_begin",
        "on_train_batch_end",
        "on_train_begin",
        "on_train_end",
        "set_model",
        "set_params",
    ]

    def __init__(self):
        self.model = None
        # Whether this Callback should only run on the chief worker in a
        # Multi-Worker setting.
        self._chief_worker_only = None

    def set_params(self, params):
        self.params = params

    def set_model(self, model):
        self.model = model

    # @doc_controls.for_subclass_implementers
    def on_epoch_begin(
        self, epoch: int, logs: tp.Optional[tp.Dict[str, jnp.ndarray]] = None
    ):
        """Calls the `on_epoch_begin` methods of its callbacks.
        This function should only be called during TRAIN mode.
        Args:
            epoch: Integer, index of epoch.
            logs: Dict. Currently no data is passed to this argument for this method
              but that may change in the future.
        """
        pass

    # @doc_controls.for_subclass_implementers
    def on_epoch_end(
        self, epoch: int, logs: tp.Optional[tp.Dict[str, jnp.ndarray]] = None
    ):
        """Called at the end of an epoch.

        Subclasses should override for any actions to run. This function should only
        be called during TRAIN mode.

        Arguments:
            epoch: integer, index of epoch.
            logs: dict, metric results for this training epoch, and for the
                validation epoch if validation is performed. Validation result keys
                are prefixed with `val_`.
        """
        pass

    # @doc_controls.for_subclass_implementers
    @default
    def on_train_batch_begin(
        self, batch: int, logs: tp.Optional[tp.Dict[str, jnp.ndarray]] = None
    ):
        """Called at the beginning of a training batch in `fit` methods.

        Subclasses should override for any actions to run.

        Arguments:
            batch: integer, index of batch within the current epoch.
            logs: dict. Has keys `batch` and `size` representing the current batch
                number and the size of the batch.
        """
        pass

    # @doc_controls.for_subclass_implementers
    @default
    def on_train_batch_end(
        self, batch: int, logs: tp.Optional[tp.Dict[str, jnp.ndarray]] = None
    ):
        """Called at the end of a training batch in `fit` methods.

        Subclasses should override for any actions to run.

        Arguments:
            batch: integer, index of batch within the current epoch.
            logs: dict. Metric results for this batch.
        """
        pass

    # @doc_controls.for_subclass_implementers
    @default
    def on_test_batch_begin(
        self, batch: int, logs: tp.Optional[tp.Dict[str, jnp.ndarray]] = None
    ):
        """Called at the beginning of a batch in `evaluate` methods.

        Also called at the beginning of a validation batch in the `fit`
        methods, if validation data is provided.

        Subclasses should override for any actions to run.

        Arguments:
            batch: integer, index of batch within the current epoch.
            logs: dict. Has keys `batch` and `size` representing the current batch
                number and the size of the batch.
        """
        pass

    # @doc_controls.for_subclass_implementers
    @default
    def on_test_batch_end(
        self, batch: int, logs: tp.Optional[tp.Dict[str, jnp.ndarray]] = None
    ):
        """Called at the end of a batch in `evaluate` methods.

        Also called at the end of a validation batch in the `fit`
        methods, if validation data is provided.

        Subclasses should override for any actions to run.

        Arguments:
            batch: integer, index of batch within the current epoch.
            logs: dict. Metric results for this batch.
        """
        pass

    # @doc_controls.for_subclass_implementers
    @default
    def on_predict_batch_begin(
        self, batch: int, logs: tp.Optional[tp.Dict[str, jnp.ndarray]] = None
    ):
        """Called at the beginning of a batch in `predict` methods.

        Subclasses should override for any actions to run.

        Arguments:
            batch: integer, index of batch within the current epoch.
            logs: dict. Has keys `batch` and `size` representing the current batch
                number and the size of the batch.
        """
        pass

    # @doc_controls.for_subclass_implementers
    @default
    def on_predict_batch_end(
        self, batch: int, logs: tp.Optional[tp.Dict[str, jnp.ndarray]] = None
    ):
        """Called at the end of a batch in `predict` methods.

        Subclasses should override for any actions to run.

        Arguments:
            batch: integer, index of batch within the current epoch.
            logs: dict. Metric results for this batch.
        """
        pass

    # @doc_controls.for_subclass_implementers
    def on_train_begin(self, logs: tp.Optional[tp.Dict[str, jnp.ndarray]] = None):
        """Called at the beginning of training.

        Subclasses should override for any actions to run.

        Arguments:
            logs: dict. Currently no data is passed to this argument for this method
                but that may change in the future.
        """
        pass

    # @doc_controls.for_subclass_implementers
    def on_train_end(self, logs: tp.Optional[tp.Dict[str, jnp.ndarray]] = None):
        """Called at the end of training.

        Subclasses should override for any actions to run.

        Arguments:
            logs: dict. Currently no data is passed to this argument for this method
                but that may change in the future.
        """
        pass

    # @doc_controls.for_subclass_implementers
    def on_test_begin(self, logs: tp.Optional[tp.Dict[str, jnp.ndarray]] = None):
        """Called at the beginning of evaluation or validation.

        Subclasses should override for any actions to run.

        Arguments:
            logs: dict. Currently no data is passed to this argument for this method
                but that may change in the future.
        """
        pass

    # @doc_controls.for_subclass_implementers
    def on_test_end(self, logs: tp.Optional[tp.Dict[str, jnp.ndarray]] = None):
        """Called at the end of evaluation or validation.

        Subclasses should override for any actions to run.

        Arguments:
            logs: dict. Currently no data is passed to this argument for this method
                but that may change in the future.
        """
        pass

    # @doc_controls.for_subclass_implementers
    def on_predict_begin(self, logs: tp.Optional[tp.Dict[str, jnp.ndarray]] = None):
        """Called at the beginning of prediction.

        Subclasses should override for any actions to run.

        Arguments:
            logs: dict. Currently no data is passed to this argument for this method
                but that may change in the future.
        """
        pass

    # @doc_controls.for_subclass_implementers
    def on_predict_end(self, logs: tp.Optional[tp.Dict[str, jnp.ndarray]] = None):
        """Called at the end of prediction.

        Subclasses should override for any actions to run.

        Arguments:
            logs: dict. Currently no data is passed to this argument for this method
                but that may change in the future.
        """
        pass

    def _implements_train_batch_hooks(self):
        """Determines if this Callback should be called for each train batch."""
        return not is_default(self.on_train_batch_begin) or not is_default(
            self.on_train_batch_end
        )

    def _implements_test_batch_hooks(self):
        """Determines if this Callback should be called for each test batch."""
        return not is_default(self.on_test_batch_begin) or not is_default(
            self.on_test_batch_end
        )

    def _implements_predict_batch_hooks(self):
        """Determines if this Callback should be called for each predict batch."""
        return not is_default(self.on_predict_batch_begin) or not is_default(
            self.on_predict_batch_end
        )



import typing as tp
from typing import Callable

import jax.numpy as jnp

from elegy.callbacks.callback import Callback

"""
Methods:
--------
on_epoch_begin
on_epoch_end
on_train_batch_begin
on_train_batch_end
on_test_batch_begin
on_test_batch_end
on_predict_batch_begin
on_predict_batch_end
on_train_begin
on_train_end
on_test_begin
on_test_end
on_predict_begin
on_predict_end
"""


class ModuleCallback(Callback):
    def on_epoch_begin(
        self, epoch: int, logs: tp.Optional[tp.Dict[str, jnp.ndarray]] = None
    ):
        assert self.model is not None
        self.model.module = self.model.module.on_epoch_begin(epoch, logs)

    def on_epoch_end(
        self, epoch: int, logs: tp.Optional[tp.Dict[str, jnp.ndarray]] = None
    ):
        assert self.model is not None
        self.model.module = self.model.module.on_epoch_end(epoch, logs)

    def on_train_batch_begin(
        self, batch: int, logs: tp.Optional[tp.Dict[str, jnp.ndarray]] = None
    ):
        assert self.model is not None
        self.model.module = self.model.module.on_train_batch_begin(batch, logs)

    def on_train_batch_end(
        self, batch: int, logs: tp.Optional[tp.Dict[str, jnp.ndarray]] = None
    ):
        assert self.model is not None
        self.model.module = self.model.module.on_train_batch_end(batch, logs)

    def on_test_batch_begin(
        self, batch: int, logs: tp.Optional[tp.Dict[str, jnp.ndarray]] = None
    ):
        assert self.model is not None
        self.model.module = self.model.module.on_test_batch_begin(batch, logs)

    def on_test_batch_end(
        self, batch: int, logs: tp.Optional[tp.Dict[str, jnp.ndarray]] = None
    ):
        assert self.model is not None
        self.model.module = self.model.module.on_test_batch_end(batch, logs)

    def on_predict_batch_begin(
        self, batch: int, logs: tp.Optional[tp.Dict[str, jnp.ndarray]] = None
    ):
        assert self.model is not None
        self.model.module = self.model.module.on_predict_batch_begin(batch, logs)

    def on_predict_batch_end(
        self, batch: int, logs: tp.Optional[tp.Dict[str, jnp.ndarray]] = None
    ):
        assert self.model is not None
        self.model.module = self.model.module.on_predict_batch_end(batch, logs)

    def on_train_begin(self, logs: tp.Optional[tp.Dict[str, jnp.ndarray]] = None):
        assert self.model is not None
        self.model.module = self.model.module.on_train_begin(logs)

    def on_train_end(self, logs: tp.Optional[tp.Dict[str, jnp.ndarray]] = None):
        assert self.model is not None
        self.model.module = self.model.module.on_train_end(logs)

    def on_test_begin(self, logs: tp.Optional[tp.Dict[str, jnp.ndarray]] = None):
        assert self.model is not None
        self.model.module = self.model.module.on_test_begin(logs)

    def on_test_end(self, logs: tp.Optional[tp.Dict[str, jnp.ndarray]] = None):
        assert self.model is not None
        self.model.module = self.model.module.on_test_end(logs)

    def on_predict_begin(self, logs: tp.Optional[tp.Dict[str, jnp.ndarray]] = None):
        assert self.model is not None
        self.model.module = self.model.module.on_predict_begin(logs)

    def on_predict_end(self, logs: tp.Optional[tp.Dict[str, jnp.ndarray]] = None):
        assert self.model is not None
        self.model.module = self.model.module.on_predict_end(logs)



# Implementation based on tf.keras.callbacks.py and elegy.callbacks.TensorBoard
# https://github.com/tensorflow/tensorflow/blob/v2.2.0/tensorflow/python/keras/callbacks.py
# https://github.com/poets-ai/elegy/blob/master/elegy/callbacks/tensorboard.py


import math
from typing import Dict, Optional, Union

from .callback import Callback


class WandbCallback(Callback):
    """
    Callback that streams epoch results to a [Weights & Biases](https://self.wandb.ai/) run.

    ```python
    self.wandb.login()
    wandb_logger = WandbCallback(project="sample-self.wandb-project", job_type="train")
    model.fit(X_train, Y_train, callbacks=[wandb_logger])
    ```
    """

    def __init__(
        self,
        project: Optional[str] = None,
        name: Optional[str] = None,
        entity: Optional[str] = None,
        job_type: Optional[str] = None,
        config: Union[Dict, str, None] = None,
        update_freq: Union[str, int] = "epoch",
        save_model: bool = False,
        monitor: str = "val_loss",
        mode: str = "min",
        **kwargs,
    ):
        """
        Arguments:
            project: (str, optional) The name of the project where you're sending the new run.
                If the project is not specified, the run is put in an "Uncategorized" project.
            name: (str, optional) A short display name for this run, which is how you'll
                identify this run in the UI. By default we generate a random two-word name that
                lets you easily cross-reference runs from the table to charts. Keeping these run
                names short makes the chart legends and tables easier to read. If you're looking
                for a place to save your hyperparameters, we recommend saving those in config.
            entity: (str, optional) An entity is a username or team name where you're sending runs.
                This entity must exist before you can send runs there, so make sure to create your
                account or team in the UI before starting to log runs. If you don't specify an entity,
                the run will be sent to your default entity, which is usually your username.
            job_type: (str, optional) Specify the type of run, which is useful when you're grouping
                runs together into larger experiments using group. For example, you might have multiple
                jobs in a group, with job types like train and eval. Setting this makes it easy to
                filter and group similar runs together in the UI so you can compare apples to apples.
            config: (dict, argparse, absl.flags, str, optional) This sets `self.wandb.config`, a dictionary-like
                object for saving inputs to your job, like hyperparameters for a model or settings for a
                data preprocessing job. The config will show up in a table in the UI that you can use to
                group, filter, and sort runs. Keys should not contain . in their names, and values should
                be under 10 MB. If dict, argparse or `absl.flags`: will load the key value pairs into the
                self.wandb.config object. If str: will look for a yaml file by that name, and load config from
                that file into the `self.wandb.config` object.
            update_freq: (str, int)`'batch'` or `'epoch'` or integer. When using `'batch'`, writes the
                losses and metrics to TensorBoard after each batch. The same applies for `'epoch'`. If
                using an integer, let's say `1000`, the callback will write the metrics and losses to
                TensorBoard every 1000 batches. Note that writing too frequently to TensorBoard can slow
                down your training.
            save_model: (bool) Save a model when monitor beats all previous epochs if set to `True` otherwise
                don't save models.
            monitor: (str) name of metric to monitor. Defaults to `'val_loss'`.
            mode: (str) one of {`'min'`, `'max'`, `'every'`}. `'min'` - save model when monitor is minimized.
                `'max'` - save model when monitor is maximized. `'every'` - save model after every epoch.
        """
        import wandb

        self.wandb = wandb
        super().__init__()
        self.run = (
            self.wandb.init(
                project=project,
                name=name,
                entity=entity,
                job_type=job_type,
                config=config,
                **kwargs,
            )
            if self.wandb.run is None
            else self.wandb.run
        )
        self.keys = None
        self.write_per_batch = True
        self._constant_fields = ["size"]
        self._constants = {}
        self._save_model = save_model
        self._monitor = monitor
        self._mode = mode
        self._monitor_metric_val = math.inf if mode == "min" else -math.inf
        self._best_epoch = 0
        self._model_path = f"model-best-0"
        self._model_checkpoint = self.model
        try:
            self.update_freq = int(update_freq)
        except ValueError as e:
            self.update_freq = 1
            if update_freq == "batch":
                self.write_per_batch = True
            elif update_freq == "epoch":
                self.write_per_batch = False
            else:
                raise e

    def _gather_configs(self):

        module_attributes = vars(vars(self.model)["module"])
        for _var in module_attributes:
            if (
                type(module_attributes[_var]) == str
                or type(module_attributes[_var]) == int
            ):
                self.wandb.run.config[_var] = module_attributes[_var]

    def _add_model_as_artifact(self, model_path: str, epoch: int):

        artifact = self.wandb.Artifact(
            f"model-{self.run.name}",
            type="model",
            metadata={"epoch": epoch, "model_path": model_path},
        )
        artifact.add_dir(model_path)
        self.run.log_artifact(artifact)

    def on_train_begin(self, logs=None):
        self.steps = self.params["steps"]
        self.global_step = 0
        self._gather_configs()

    def on_train_batch_end(self, batch: int, logs=None):
        if not self.write_per_batch:
            return
        logs = logs or {}
        for key in self._constant_fields:
            self._constants[key] = logs[key]
            logs.pop(key, None)
            logs.pop("val_" + key, None)
        self.global_step = batch + self.current_epoch * (self.steps)
        if self.global_step % self.update_freq == 0:
            if self.keys is None:
                self.keys = logs.keys()
            for key in self.keys:
                log_key = key
                if log_key[:4] != "val_":
                    log_key = "train_" + log_key
                self.run.log({log_key: logs[key]}, step=self.global_step)

    def on_epoch_begin(self, epoch: int, logs=None):
        self.current_epoch = epoch

    def on_epoch_end(self, epoch: int, logs=None):
        logs = logs or {}
        for key in self._constant_fields:
            self._constants[key] = logs[key]
            logs.pop(key, None)
            logs.pop("val_" + key, None)

        if self.keys is None:
            self.keys = logs.keys()

        if self.write_per_batch:
            for key in logs:
                log_key = key
                if log_key[:4] != "val_":
                    log_key = "train_" + log_key
                self.run.log({log_key: logs[key]}, step=self.global_step)

        elif epoch % self.update_freq == 0:
            for key in logs:
                log_key = key
                if log_key[:4] != "val_":
                    log_key = "train_" + log_key
                self.run.log({log_key: logs[key]}, step=epoch)

        if self._save_model:
            if self._mode == "every":
                self._best_epoch = epoch
                self._model_path = f"model-{epoch + 1}-{self.run.name}"
                print(f"Saving Model at {self._model_path}")
                self._model_checkpoint = self.model
                self._model_checkpoint.save(self._model_path)

            elif self._mode == "min" and logs[self._monitor] < self._monitor_metric_val:
                self._best_epoch = epoch
                self._model_path = f"model-best-{epoch + 1}-{self.run.name}"
                print(
                    f"{self._monitor} decreased at epoch {epoch}. Saving Model at {self._model_path}"
                )
                self._model_checkpoint = self.model
                self._model_checkpoint.save(self._model_path)
                self._monitor_metric_val = logs[self._monitor]

            elif self._mode == "max" and logs[self._monitor] > self._monitor_metric_val:
                self._best_epoch = epoch
                self._model_path = f"model-best-{epoch + 1}-{self.run.name}"
                print(
                    f"{self._monitor} increased at epoch {epoch}. Saving Model at {self._model_path}"
                )
                self._model_checkpoint = self.model
                self._model_checkpoint.save(self._model_path)
                self._monitor_metric_val = logs[self._monitor]

            self._add_model_as_artifact(self._model_path, epoch=self._best_epoch)

    def on_train_end(self, logs=None):

        for key in self._constant_fields:
            self.wandb.run.summary[key] = self._constants[key]
        self.run.finish()



import typing as tp

import flax
import jax
import jax.numpy as jnp
import jax_metrics as jm
import numpy as np
import treex as tx
from optax import GradientTransformation

from elegy import types, utils
from elegy.model.model_base import ModelBase
from elegy.model.model_core import (
    GradStepOutput,
    PredStepOutput,
    TestStepOutput,
    TrainStepOutput,
)

M = tp.TypeVar("M", bound="Model")
U = tp.TypeVar("U", bound="tx.Module")

try:
    import haiku as hk

    TransformedWithState = hk.TransformedWithState
    HaikuModule = tx.HaikuModule
except (ImportError, ModuleNotFoundError):
    hk = None
    TransformedWithState = type(None)
    HaikuModule = tp.cast(tp.Any, None)


class Model(tp.Generic[U], ModelBase):
    """
    Model provides an Estimator-like API similar to Keras.
    """

    # pytree
    module: tp.Optional[U] = tx.node()
    loss_and_logs: tp.Optional[jm.LossesAndMetrics]
    optimizer: tp.Optional[tx.Optimizer]

    # static
    seed: int = 42

    @tp.overload
    def __init__(
        self: "Model[tx.FlaxModule]",
        module: flax.linen.module.Module,
        loss: tp.Any = None,
        metrics: tp.Any = None,
        optimizer: tp.Optional[tp.Union[tx.Optimizer, GradientTransformation]] = None,
        seed: int = 42,
        eager: bool = False,
    ):
        ...

    if hk is not None:

        @tp.overload
        def __init__(
            self: "Model[HaikuModule]",
            module: hk.TransformedWithState,
            loss: tp.Any = None,
            metrics: tp.Any = None,
            optimizer: tp.Optional[
                tp.Union[tx.Optimizer, GradientTransformation]
            ] = None,
            seed: int = 42,
            eager: bool = False,
        ):
            ...

    @tp.overload
    def __init__(
        self: "Model[U]",
        module: U,
        loss: tp.Any = None,
        metrics: tp.Any = None,
        optimizer: tp.Optional[tp.Union[tx.Optimizer, GradientTransformation]] = None,
        seed: int = 42,
        eager: bool = False,
    ):
        ...

    def __init__(
        self,
        module: tp.Union[
            flax.linen.module.Module, TransformedWithState, tx.Module, None
        ] = None,
        loss: tp.Any = None,
        metrics: tp.Any = None,
        optimizer: tp.Optional[tp.Union[tx.Optimizer, GradientTransformation]] = None,
        seed: int = 42,
        eager: bool = False,
    ):
        """
        Arguments:
            module: A `Module` instance.
            loss: A `elegy.Loss` or `Callable` instance representing the loss function of the network.
                You can define more loss terms by simply passing a possibly nested structure of
                lists and dictionaries of `elegy.Loss` or `Callable`s. Usually a plain list of losses is enough
                but using dictionaries will create namescopes for the names of the losses
                which might be useful e.g. to group things in tensorboard. Contrary to Keras convention,
                in Elegy there is no relation between the structure of `loss` with the structure
                of the labels and outputs of the network. Elegy's loss system is more flexible than
                the one provided by Keras, for more information on how to mimick Keras behavior checkout the
                [Losses and Metrics Guide](https://poets-ai.github.io/elegy/guides/losses-and-metrics)`.
            metrics: A `elegy.Metric` or `Callable` instance representing the loss function of the network.
                You can define more metrics terms by simply passing a possibly nested structure of
                lists and dictionaries of `elegy.Metric` or `Callable`s. Usually a plain list of metrics is enough
                but using dictionaries will create namescopes for the names of the metrics
                which might be useful e.g. to group things in tensorboard. Contrary to Keras convention,
                in Elegy there is no relation between the structure of `metrics` with the structure
                of the labels and outputs of the network. Elegy's metrics system is more flexible than
                the one provided by Keras, for more information on how to mimick Keras behavior checkout the
                [Losses and Metrics Guide](https://poets-ai.github.io/elegy/guides/losses-and-metrics)`.
            optimizer: A `optax` optimizer instance. Optix is a very flexible library for defining
                optimization pipelines with things like learning rate schedules, this means that
                there is no need for a `LearningRateScheduler` callback in Elegy.
            eager: Settable attribute indicating whether the model should run eagerly.
                Running eagerly means that your model will be run step by step, like Python code, instead of
                using Jax's `jit` to. Your model might run slower, but it should become easier for you to debug
                it by stepping into individual layer calls.
        """
        super().__init__(seed=seed, eager=eager)

        if isinstance(module, flax.linen.module.Module):
            self.module = tx.FlaxModule(module)

        elif TransformedWithState is not None and isinstance(
            module, TransformedWithState
        ):
            self.module = HaikuModule(module)
        else:
            self.module = module

        self.optimizer = (
            tx.Optimizer(optimizer)
            if isinstance(optimizer, GradientTransformation)
            else optimizer
        )
        self.loss_and_logs = None

        self._losses_and_metrics = tx.Hashable((loss, metrics))

    def __call__(self, *args, **kwargs) -> tp.Any:
        assert self.module is not None

        return self.module(*args, **kwargs)

    # ----------------------------------------------------------------
    # implement low-level API methods
    # ----------------------------------------------------------------

    def init_step(
        self: M,
        key: jnp.ndarray,
        inputs: tp.Any,
    ) -> M:
        model: M = self

        if model.module is not None:
            model.module = model.module.init(key, inputs)

        if model.optimizer is not None:
            params = model.parameters()
            model.optimizer = model.optimizer.init(params)

        losses, metrics = model._losses_and_metrics.value
        aux_losses = model.loss_logs()
        aux_metrics = model.metric_logs()

        model.loss_and_logs = jm.LossesAndMetrics(
            losses=losses,
            metrics=metrics,
            aux_losses=aux_losses,
            aux_metrics=aux_metrics,
        )

        model = model.distributed_strategy.handle_post_init(model)

        return model

    def pred_step(
        self: M,
        inputs: tp.Any,
    ) -> PredStepOutput[M]:
        model: M = self

        if model.module is None:
            raise types.MissingModule(
                "Trying to run default `pred_step` on a Model with no `module`, try overriding `pred_step` or set `module`"
            )

        inputs_obj = tx.Inputs.from_value(inputs)

        preds = model.module(*inputs_obj.args, **inputs_obj.kwargs)

        return preds, model

    def test_step(
        self: M,
        inputs: tp.Any,
        labels: tp.Mapping[str, tp.Any],
    ) -> TestStepOutput[M]:
        model: M = self

        if model.module is None:
            raise types.MissingModule(
                "Trying to run default `test_step` on a Model with no `loss_and_logs`, try overriding `test_step` or set `loss_and_logs`"
            )

        preds, model = model.pred_step(inputs)
        assert model.loss_and_logs is not None

        aux_losses = model.loss_logs()
        aux_metrics = model.metric_logs()

        extended_labels = {
            "inputs": inputs,
            "preds": preds,
            "model": model,
            "parameters": model.parameters(),
            "batch_stats": model.batch_stats(),
            "rngs": model.rngs(),
            "model_states": model.model_states(),
            "states": model.states(),
            "metric_logs": model.metric_logs(),
            "loss_logs": model.loss_logs(),
            "logs": model.logs(),
            **labels,
        }

        losses_kwargs = extended_labels
        metrics_kwargs = extended_labels

        # loss, *_ = model.loss_and_logs(
        #     **losses_kwargs,
        #     metrics_kwargs=metrics_kwargs,
        #     aux_losses=aux_losses,
        #     aux_metrics=aux_metrics,
        # )

        # TODO: could be refactored as reset_update in treex?
        batch_loss_and_logs = model.loss_and_logs.copy()
        batch_loss_and_logs.reset()
        batch_loss_and_logs.update(
            **losses_kwargs,
            metrics_kwargs=metrics_kwargs,
            aux_losses=aux_losses,
            aux_metrics=aux_metrics,
        )
        loss, *_ = batch_loss_and_logs.compute()

        batch_loss_and_logs = model.distributed_strategy.handle_metrics(
            batch_loss_and_logs
        )

        self.loss_and_logs = jax.tree_map(
            lambda a, b: a + b, self.loss_and_logs, batch_loss_and_logs
        )

        _, losses_logs, metrics_logs = model.loss_and_logs.compute()

        logs = {**losses_logs, **metrics_logs}

        return loss, logs, model

    @staticmethod
    def loss_fn(
        params: M,
        model: M,
        inputs: tp.Any,
        labels: tp.Mapping[str, tp.Any],
    ) -> tp.Tuple[jnp.ndarray, tp.Tuple[types.Logs, M]]:
        model = model.merge(params)
        loss, logs, model = model.test_step(inputs, labels)
        return loss, (logs, model)

    @staticmethod
    def _is_trainable(field_info: tx.FieldInfo) -> bool:
        if isinstance(field_info.module, tx.Module):
            return not field_info.module.frozen and field_info.module.training
        else:
            return True

    def train_step(
        self: M,
        inputs: tp.Any,
        labels: tp.Mapping[str, tp.Any],
    ) -> TrainStepOutput[M]:
        model: M = self
        grads: M
        logs: types.Logs
        model: M = self

        if model.optimizer is None:
            raise types.MissingModule(
                "Trying to run default `train_step` on a Model with no `optimizer`, try overriding `train_step` or set `optimizer`"
            )

        params = model.parameters(self._is_trainable)

        grad_fn = jax.grad(self.loss_fn, has_aux=True)
        grads, (logs, model) = grad_fn(params, model, inputs, labels)

        model, grads = model.distributed_strategy.handle_model_and_grads(model, grads)

        assert model.optimizer is not None

        params = model.optimizer.update(grads, params)
        model = model.merge(params)

        return logs, model

    def reset_step(self) -> None:
        self.module = self.module.reset_step()

    # ----------------------------------------------------------------
    # Model-only methods
    # ----------------------------------------------------------------

    def summary(
        self,
        inputs: tp.Any = tx.MISSING,
        depth: int = 2,
        return_repr: bool = False,
    ) -> tp.Optional[str]:
        """
        Prints a summary of the network. The representation is module dependent,
        if a library provides a representation, it will be used, else elegy will
        define its own.

        Arguments:
            x: A sample of inputs to the network.
            depth: The level number of nested level which will be showed.
                Information about summaries from modules deeper than `depth`
                will be aggregated together.
            return_repr: If True, the summary will be returned as a string and will not be printed.
            eval_shape: If True, jax.eval_shape is used to calculate all shapes, this avoids actually
                running the computation as only shapes are calculated (turn off if trying to debug).
        """
        # model = self.local()

        # assert model.module is not None

        # if not model.initialized:
        #     if inputs is tx.MISSING:
        #         raise ValueError(
        #             "`inputs` is required to print the summary of uninitialized Models"
        #         )

        #     model.init_on_batch(inputs)

        model = self.local()

        assert model.module is not None

        if not model.initialized:
            if inputs is tx.MISSING:
                raise ValueError(
                    "`inputs` is required to print the summary of uninitialized Models"
                )

            model.init_on_batch(inputs)

        summary = model.module.tabulate(
            inputs=inputs,
            depth=depth,
        )

        if return_repr:
            return summary
        else:
            print(summary)



import typing as tp
from dataclasses import dataclass

import jax
import jax.numpy as jnp
import treex as tx
from einop import einop

import elegy.modules.core_module as cm


@dataclass(unsafe_hash=True)
class Eager(cm.DistributedStrategy):
    def init_step_fn(self, model: cm.M) -> cm.InitStep[cm.M]:
        return model.__class__.init_step

    def pred_step_fn(self, model: cm.M) -> cm.PredStep[cm.M]:
        return model.__class__.predict_step

    def test_step_fn(self, model: cm.M) -> cm.TestStep[cm.M]:
        return model.__class__.test_step

    def train_step_fn(self, model: cm.M) -> cm.TrainStep[cm.M]:
        return model.__class__.train_step


@dataclass(unsafe_hash=True)
class JIT(cm.DistributedStrategy):
    # donate 'model' memory buffer since we return an updated model
    def init_step_fn(self, model: cm.M) -> cm.InitStep[cm.M]:
        return jax.jit(model.__class__.init_step, donate_argnums=0)

    def pred_step_fn(self, model: cm.M) -> cm.PredStep[cm.M]:
        return jax.jit(model.__class__.predict_step, donate_argnums=0)

    def test_step_fn(self, model: cm.M) -> cm.TestStep[cm.M]:
        return jax.jit(model.__class__.test_step, donate_argnums=0)

    def train_step_fn(self, model: cm.M) -> cm.TrainStep[cm.M]:
        return jax.jit(model.__class__.train_step, donate_argnums=0)


@dataclass(unsafe_hash=True)
class DataParallel(cm.DistributedStrategy):
    def from_local(self, model: cm.M) -> cm.M:
        # device_idxs used to inform pmap about the number of devices
        device_idxs = jnp.arange(jax.device_count())
        model = jax.pmap(
            lambda idx, model: model,
            in_axes=(0, None),
            out_axes=0,
        )(device_idxs, model)

        return model

    def to_local(self, model: cm.M) -> cm.M:
        return jax.tree_map(lambda x: x[0], model)

    def lift_data(self, data: cm.A) -> cm.A:
        data = jax.tree_map(
            lambda x: einop(
                x,
                "(device batch) ... -> device batch ...",
                device=jax.device_count(),
            ),
            data,
        )
        return data

    def lift_key(self, key: jnp.ndarray) -> jnp.ndarray:
        key = einop(
            key,
            "... -> device ...",
            device=jax.device_count(),
        )
        return key

    def lift_batch_size(self, batch_size: int) -> int:
        return batch_size * jax.device_count()

    def handle_post_init(self, model: cm.M) -> cm.M:
        return model.map(
            lambda key: jax.random.fold_in(key, jax.lax.axis_index("device")),
            tx.Rng,
            inplace=True,
        )

    def handle_metrics(
        self,
        metrics: cm.Me,
    ) -> cm.Me:
        metrics = jax.lax.stop_gradient(metrics)
        metrics = jax.lax.psum(metrics, axis_name="device")
        return metrics

    def handle_model_and_grads(
        self,
        model: cm.M,
        grads: cm.M,
    ) -> tp.Tuple[cm.M, cm.M]:
        grads = jax.lax.pmean(grads, axis_name="device")
        model = model.map(lambda x: jax.lax.pmean(x, axis_name="device"), tx.BatchStat)

        return model, grads

    def init_step_fn(self, model: cm.M) -> cm.InitStep[cm.M]:
        return jax.pmap(
            model.__class__.init_step,
            axis_name="device",
            donate_argnums=0,
        )

    def pred_step_fn(self, model: cm.M) -> cm.PredStep[cm.M]:
        return jax.pmap(
            model.__class__.predict_step,
            axis_name="device",
            donate_argnums=0,
        )

    def test_step_fn(self, model: cm.M) -> cm.TestStep[cm.M]:
        return jax.pmap(
            model.__class__.test_step,
            axis_name="device",
            out_axes=(0, None, 0),  # None = logs not replicated
            donate_argnums=0,
        )

    def train_step_fn(self, model: cm.M) -> cm.TrainStep[cm.M]:
        return jax.pmap(
            model.__class__.train_step,
            axis_name="device",
            out_axes=(None, 0),  # None = logs not replicated
            donate_argnums=0,
        )



# # adapted from the flax library https://github.com/google/flax

# import pickle
# import typing as tp

# import jax
# import jax.numpy as jnp
# import numpy as np
# from elegy import hooks, module, nn, types, utils

# __all__ = [
#     "ResNet",
#     "ResNet18",
#     "ResNet34",
#     "ResNet50",
#     "ResNet101",
#     "ResNet152",
#     "ResNet200",
# ]


# PRETRAINED_URLS = {
#     "ResNet18": {
#         "url": "https://github.com/poets-ai/elegy-assets/releases/download/resnet18_rev1/ResNet18_ImageNet_rev1.pkl",
#         "sha256": "02824ae2f29563add46feff14f40c362ae5f9af3f01ea2edc0812e5ca06ca9ae",
#     },
#     "ResNet50": {
#         "url": "https://github.com/poets-ai/elegy-assets/releases/download/resnet50_rev1/ResNet50_ImageNet_rev1.pkl",
#         "sha256": "c69086813ccff6b67b2452daabdf64772f8a7f5c04591e1962185129e18989fc",
#     },
# }


# class ResNetBlock(module.Module):
#     """ResNet (identity) block"""

#     def __init__(
#         self,
#         n_filters: int,
#         strides: tp.Optional[tp.Tuple[int]] = (1, 1),
#         *args,
#         **kwargs,
#     ):
#         super().__init__(*args, **kwargs)
#         self.n_filters = n_filters
#         self.strides = strides

#     def __call__(self, x: jnp.ndarray):
#         x0 = x
#         x = nn.Conv(
#             self.n_filters,
#             (3, 3),
#             with_bias=False,
#             stride=self.strides,
#             dtype=self.dtype,
#         )(x)
#         x = nn.BatchNorm(decay_rate=0.9, eps=1e-5)(x)
#         x = jax.nn.relu(x)

#         x = nn.Conv(self.n_filters, (3, 3), with_bias=False, dtype=self.dtype)(x)
#         x = nn.BatchNorm(decay_rate=0.9, eps=1e-5)(x)

#         if x0.shape != x.shape:
#             x0 = nn.Conv(
#                 self.n_filters,
#                 (1, 1),
#                 with_bias=False,
#                 stride=self.strides,
#                 dtype=self.dtype,
#             )(x0)
#             x0 = nn.BatchNorm(decay_rate=0.9, eps=1e-5)(x0)
#         return jax.nn.relu(x0 + x)


# class BottleneckResNetBlock(ResNetBlock):
#     """ResNet Bottleneck block."""

#     def __call__(self, x: jnp.ndarray):
#         x0 = x
#         x = nn.Conv(self.n_filters, (1, 1), with_bias=False, dtype=self.dtype)(x)
#         x = nn.BatchNorm(decay_rate=0.9, eps=1e-5)(x)
#         x = jax.nn.relu(x)
#         x = nn.Conv(
#             self.n_filters,
#             (3, 3),
#             with_bias=False,
#             stride=self.strides,
#             dtype=self.dtype,
#         )(x)
#         x = nn.BatchNorm(decay_rate=0.9, eps=1e-5)(x)
#         x = jax.nn.relu(x)
#         x = nn.Conv(self.n_filters * 4, (1, 1), with_bias=False, dtype=self.dtype)(x)
#         x = nn.BatchNorm(decay_rate=0.9, eps=1e-5, scale_init=jnp.zeros)(x)

#         if x0.shape != x.shape:
#             x0 = nn.Conv(
#                 self.n_filters * 4,
#                 (1, 1),
#                 with_bias=False,
#                 stride=self.strides,
#                 dtype=self.dtype,
#             )(x0)
#             x0 = nn.BatchNorm(decay_rate=0.9, eps=1e-5)(x0)
#         return jax.nn.relu(x0 + x)


# class ResNet(module.Module):
#     """A generic ResNet V1 architecture that can be customized for non-standard configurations
#     Original Paper: [Deep Residual Learning for Image Recognition](https://arxiv.org/abs/1512.03385)
#     """

#     __all__ = ["__init__", "call"]

#     def __init__(
#         self,
#         stages: tp.List[int],
#         block_type: tp.Union[tp.Type[ResNetBlock], tp.Type[BottleneckResNetBlock]],
#         lowres: bool = False,
#         weights: tp.Optional[str] = None,
#         dtype: tp.Optional[tp.Any] = jnp.float32,
#         *args,
#         **kwargs,
#     ):
#         """
#         Arguments:
#             stages: A list of integers representing the number of blocks in each stage.
#                     e.g: [3, 4, 6, 3] for a ResNet50
#             block_type: Which ResNet block type to use.
#             lowres: Optional, whether to use the low resolution version
#                     as described in subsection 4.2 of the orignal paper.
#                     This version is better suited for datasets like CIFAR10. (Default: False)
#             weights: One of None (random initialization) or a path to a weights file
#             dtype: Optional dtype of the convolutions and linear operations,
#                     either jnp.float32 (default) or jnp.float16 for mixed precision.
#         """

#         super().__init__(*args, **kwargs)
#         self.stages = stages
#         self.block_type = block_type
#         self.lowres = lowres

#         if weights is not None:
#             if weights.endswith(".pkl"):
#                 collections = pickle.load(open(weights, "rb"))
#             elif weights == "imagenet":
#                 clsname = self.__class__.__name__
#                 urldict = PRETRAINED_URLS.get(clsname, None)
#                 if urldict is None:
#                     raise ValueError(f"No pretrained weights for {clsname} available")
#                 fname = utils.download_file(urldict["url"], sha256=urldict["sha256"])
#                 collections = pickle.load(open(fname, "rb"))
#             else:
#                 raise ValueError("Unknown weights value: ", weights)

#             if isinstance(collections, tuple):
#                 parameters, collections = collections
#             elif "parameters" in collections:
#                 parameters = collections.pop("parameters")
#             else:
#                 raise ValueError(
#                     "Unknown parameters structure, expected either tuple (parameters, collections) or a collections dict with a 'parameters' field."
#                 )

#             x = np.empty([0, 224, 224, 3], dtype=self.dtype)
#             # quick but dirty module initialization
#             jax.eval_shape(self.init(rng=types.KeySeq(42)), x)

#             self.set_default_parameters(parameters, collections)

#     def __call__(self, x: jnp.ndarray):
#         x = nn.Conv(
#             64,
#             (7, 7) if not self.lowres else (3, 3),
#             stride=(2, 2) if not self.lowres else (1, 1),
#             padding="SAME",
#             with_bias=False,
#             dtype=self.dtype,
#         )(x)
#         x = nn.BatchNorm(decay_rate=0.9, eps=1e-5)(x)
#         x = module.compact_module(jax.nn.relu)()(x)

#         if not self.lowres:
#             x = nn.MaxPool(
#                 window_shape=(1, 3, 3, 1), strides=(1, 2, 2, 1), padding="SAME"
#             )(x)
#         for i, block_size in enumerate(self.stages):
#             for j in range(block_size):
#                 strides = (2, 2) if i > 0 and j == 0 else (1, 1)
#                 x = self.block_type(64 * 2 ** i, strides=strides, dtype=self.dtype)(x)
#         GAP = lambda x: jnp.mean(x, axis=(1, 2))
#         x = module.compact_module(GAP)(name="global_average_pooling")(x)
#         x = nn.Linear(1000, dtype=self.dtype)(x)
#         to_float32 = lambda x: jnp.asarray(x, jnp.float32)
#         x = module.compact_module(to_float32)(name="to_float32")(x)
#         return x


# class ResNet18(ResNet):
#     def __init__(
#         self,
#         lowres: bool = False,
#         weights: tp.Optional[str] = None,
#         dtype: tp.Optional[tp.Any] = jnp.float32,
#         *args,
#         **kwargs,
#     ):
#         super().__init__(
#             stages=[2, 2, 2, 2],
#             block_type=ResNetBlock,
#             lowres=lowres,
#             weights=weights,
#             dtype=dtype,
#             *args,
#             **kwargs,
#         )


# class ResNet34(ResNet):
#     def __init__(
#         self,
#         lowres: bool = False,
#         weights: tp.Optional[str] = None,
#         dtype: tp.Optional[tp.Any] = jnp.float32,
#         *args,
#         **kwargs,
#     ):
#         super().__init__(
#             stages=[3, 4, 6, 3],
#             block_type=ResNetBlock,
#             lowres=lowres,
#             weights=weights,
#             dtype=dtype,
#             *args,
#             **kwargs,
#         )


# class ResNet50(ResNet):
#     def __init__(
#         self,
#         lowres: bool = False,
#         weights: tp.Optional[str] = None,
#         dtype: tp.Optional[tp.Any] = jnp.float32,
#         *args,
#         **kwargs,
#     ):
#         super().__init__(
#             stages=[3, 4, 6, 3],
#             block_type=BottleneckResNetBlock,
#             lowres=lowres,
#             weights=weights,
#             dtype=dtype,
#             *args,
#             **kwargs,
#         )


# class ResNet101(ResNet):
#     def __init__(
#         self,
#         lowres: bool = False,
#         weights: tp.Optional[str] = None,
#         dtype: tp.Optional[tp.Any] = jnp.float32,
#         *args,
#         **kwargs,
#     ):
#         super().__init__(
#             stages=[3, 4, 23, 3],
#             block_type=BottleneckResNetBlock,
#             lowres=lowres,
#             weights=weights,
#             dtype=dtype,
#             *args,
#             **kwargs,
#         )


# class ResNet152(ResNet):
#     def __init__(
#         self,
#         lowres: bool = False,
#         weights: tp.Optional[str] = None,
#         dtype: tp.Optional[tp.Any] = jnp.float32,
#         *args,
#         **kwargs,
#     ):
#         super().__init__(
#             stages=[3, 8, 36, 3],
#             block_type=BottleneckResNetBlock,
#             lowres=lowres,
#             weights=weights,
#             dtype=dtype,
#             *args,
#             **kwargs,
#         )


# class ResNet200(ResNet):
#     def __init__(
#         self,
#         lowres: bool = False,
#         weights: tp.Optional[str] = None,
#         dtype: tp.Optional[tp.Any] = jnp.float32,
#         *args,
#         **kwargs,
#     ):
#         super().__init__(
#             stages=[3, 24, 36, 3],
#             block_type=BottleneckResNetBlock,
#             lowres=lowres,
#             weights=weights,
#             dtype=dtype,
#             *args,
#             **kwargs,
#         )


# _resnet__init___docstring = """
# Instantiates the {} architecture from [Deep Residual Learning for Image Recognition](https://arxiv.org/abs/1512.03385)

# Arguments:
#     lowres: Optional, whether to use the low resolution version
#             as described in subsection 4.2 of the orignal paper.
#             This version is better suited for datasets like CIFAR10. (Default: False)
#     weights: One of None (random initialization), 'imagenet' (automatic download of
#               weights pretrained on ImageNet) or a path to a weights file
#     dtype: Optional dtype of the convolutions and linear operations,
#            either jnp.float32 (default) or jnp.float16 for mixed precision.
# """

# ResNet18.__init__.__doc__ = _resnet__init___docstring.format("ResNet18")
# ResNet34.__init__.__doc__ = _resnet__init___docstring.format("ResNet34")
# ResNet50.__init__.__doc__ = _resnet__init___docstring.format("ResNet50")
# ResNet101.__init__.__doc__ = _resnet__init___docstring.format("ResNet101")
# ResNet152.__init__.__doc__ = _resnet__init___docstring.format("ResNet152")
# ResNet200.__init__.__doc__ = _resnet__init___docstring.format("ResNet200")



# isort:skip_file

__version__ = "0.8.6"

from treex import *

import elegy.types as types
import elegy.utils as utils


from . import (
    callbacks,
    data,
    model,
    # nets,
)

from .model.model import Model
from .model.model_base import ModelBase, load
from .model.model_core import (
    GradStepOutput,
    PredStepOutput,
    TestStepOutput,
    TrainStepOutput,
    LossStepOutput,
    ModelCore,
)
from .types import KeySeq
from .utils import inject_dependencies
from elegy.modules.core_module import CoreModule
from elegy.model import model_full



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['elegy',
 'elegy.callbacks',
 'elegy.data',
 'elegy.model',
 'elegy.modules',
 'elegy.nets']

package_data = \
{'': ['*']}

install_requires = \
['cloudpickle>=1.5.0,<2.0.0',
 'einop>=0.0.1,<0.0.2',
 'jax_metrics>=0.1.0.a2,<0.2.0',
 'tensorboardx>=2.1,<3.0',
 'treex>=0.7.0.a2,<0.8.0']

setup_kwargs = {
    'name': 'elegy',
    'version': '0.9.0.dev2',
    'description': 'Elegy is a Neural Networks framework based on Jax and Haiku.',
    'long_description': '# Elegy\n\n<!-- [![PyPI Status Badge](https://badge.fury.io/py/eg.svg)](https://pypi.org/project/elegy/) -->\n<!-- [![PyPI - Python Version](https://img.shields.io/pypi/pyversions/elegy)](https://pypi.org/project/elegy/) -->\n<!-- [![Documentation](https://img.shields.io/badge/api-reference-blue.svg)](https://poets-ai.github.io/elegy/) -->\n<!-- [![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black) -->\n[![Coverage](https://img.shields.io/codecov/c/github/poets-ai/elegy?color=%2334D058)](https://codecov.io/gh/poets-ai/elegy)\n[![Status](https://github.com/poets-ai/elegy/workflows/GitHub%20CI/badge.svg)](https://github.com/poets-ai/elegy/actions?query=workflow%3A%22GitHub+CI%22)\n[![Contributions welcome](https://img.shields.io/badge/contributions-welcome-brightgreen.svg?style=flat)](https://github.com/poets-ai/elegy/issues)\n\n______________________________________________________________________\n\n_A High Level API for Deep Learning in JAX_\n\n#### Main Features\n\n- 😀 **Easy-to-use**: Elegy provides a Keras-like high-level API that makes it very easy to use for most common tasks.\n- 💪\u200d **Flexible**: Elegy provides a Pytorch Lightning-like low-level API that offers maximum flexibility when needed.\n- 🔌 **Compatible**: Elegy supports various frameworks and data sources including Flax & Haiku Modules, Optax Optimizers, TensorFlow Datasets, Pytorch DataLoaders, and more.\n<!-- - 🤷 **Agnostic**: Elegy supports various frameworks, including Flax, Haiku, and Optax on the high-level API, and it is 100% framework-agnostic on the low-level API. -->\n\nElegy is built on top of [Treex](https://github.com/cgarciae/treex) and [Treeo](https://github.com/cgarciae/treeo) and reexports their APIs for convenience.\n\n\n[Getting Started](https://poets-ai.github.io/elegy/getting-started/high-level-api) | [Examples](/examples) | [Documentation](https://poets-ai.github.io/elegy)\n\n\n## What is included?\n* A `Model` class with an Estimator-like API.\n* A `callbacks` module with common Keras callbacks.\n\n**From Treex**\n\n* A `Module` class.\n* A `nn` module for with common layers.\n* A `losses` module with common loss functions.\n* A `metrics` module with common metrics.\n\n## Installation\n\nInstall using pip:\n\n```bash\npip install elegy\n```\n\nFor Windows users, we recommend the Windows subsystem for Linux 2 [WSL2](https://docs.microsoft.com/es-es/windows/wsl/install-win10?redirectedfrom=MSDN) since [jax](https://github.com/google/jax/issues/438) does not support it yet.\n\n## Quick Start: High-level API\n\nElegy\'s high-level API provides a straightforward interface you can use by implementing the following steps:\n\n**1.** Define the architecture inside a `Module`:\n\n```python\nimport jax\nimport elegy as eg\n\nclass MLP(eg.Module):\n    @eg.compact\n    def __call__(self, x):\n        x = eg.Linear(300)(x)\n        x = jax.nn.relu(x)\n        x = eg.Linear(10)(x)\n        return x\n```\n\n**2.** Create a `Model` from this module and specify additional things like losses, metrics, and optimizers:\n\n```python\nimport optax optax\nimport elegy as eg\n\nmodel = eg.Model(\n    module=MLP(),\n    loss=[\n        eg.losses.Crossentropy(),\n        eg.regularizers.L2(l=1e-5),\n    ],\n    metrics=eg.metrics.Accuracy(),\n    optimizer=optax.rmsprop(1e-3),\n)\n```\n\n**3.** Train the model using the `fit` method:\n\n```python\nmodel.fit(\n    inputs=X_train,\n    labels=y_train,\n    epochs=100,\n    steps_per_epoch=200,\n    batch_size=64,\n    validation_data=(X_test, y_test),\n    shuffle=True,\n    callbacks=[eg.callbacks.TensorBoard("summaries")]\n)\n```\n#### Using Flax\n\n<details>\n<summary>Show</summary>\n\nTo use Flax with Elegy just create a `flax.linen.Module` and pass it to `Model`.\n\n```python\nimport jax\nimport elegy as eg\nimport optax optax\nimport flax.linen as nn\n\nclass MLP(nn.Module):\n    @nn.compact\n    def __call__(self, x, training: bool):\n        x = nn.Dense(300)(x)\n        x = jax.nn.relu(x)\n        x = nn.Dense(10)(x)\n        return x\n\n\nmodel = eg.Model(\n    module=MLP(),\n    loss=[\n        eg.losses.Crossentropy(),\n        eg.regularizers.L2(l=1e-5),\n    ],\n    metrics=eg.metrics.Accuracy(),\n    optimizer=optax.rmsprop(1e-3),\n)\n```\n\nAs shown here, Flax Modules can optionally request a `training` argument to `__call__` which will be provided by Elegy / Treex. \n\n</details>\n\n#### Using Haiku\n\n<details>\n<summary>Show</summary>\n\nTo use Haiku with Elegy do the following: \n\n* Create a `forward` function.\n* Create a `TransformedWithState` object by feeding `forward` to `hk.transform_with_state`.\n* Pass your `TransformedWithState`  to `Model`.\n\nYou can also optionally create your own `hk.Module` and use it in `forward` if needed. Putting everything together should look like this:\n\n```python\nimport jax\nimport elegy as eg\nimport optax optax\nimport haiku as hk\n\n\ndef forward(x, training: bool):\n    x = hk.Linear(300)(x)\n    x = jax.nn.relu(x)\n    x = hk.Linear(10)(x)\n    return x\n\n\nmodel = eg.Model(\n    module=hk.transform_with_state(forward),\n    loss=[\n        eg.losses.Crossentropy(),\n        eg.regularizers.L2(l=1e-5),\n    ],\n    metrics=eg.metrics.Accuracy(),\n    optimizer=optax.rmsprop(1e-3),\n)\n```\n\nAs shown here, `forward` can optionally request a `training` argument which will be provided by Elegy / Treex. \n\n</details>\n\n## Quick Start: Low-level API\n\nElegy\'s low-level API lets you explicitly define what goes on during training, testing, and inference. Let\'s define our own custom `Model` to implement a `LinearClassifier` with pure JAX:\n\n**1.** Define a custom `init_step` method:\n\n```python\nclass LinearClassifier(eg.Model):\n    # use treex\'s API to declare parameter nodes\n    w: jnp.ndarray = eg.Parameter.node()\n    b: jnp.ndarray = eg.Parameter.node()\n\n    def init_step(self, key: jnp.ndarray, inputs: jnp.ndarray):\n        self.w = jax.random.uniform(\n            key=key,\n            shape=[features_in, 10],\n        )\n        self.b = jnp.zeros([10])\n\n        self.optimizer = self.optimizer.init(self)\n\n        return self\n```\nHere we declared the parameters `w` and `b` using Treex\'s `Parameter.node()` for pedagogical reasons, however normally you don\'t have to do this since you typically use a sub-`Module` instead.\n\n**2.** Define a custom `test_step` method:\n```python\n    def test_step(self, inputs, labels):\n        # flatten + scale\n        inputs = jnp.reshape(inputs, (inputs.shape[0], -1)) / 255\n\n        # forward\n        logits = jnp.dot(inputs, self.w) + self.b\n\n        # crossentropy loss\n        target = jax.nn.one_hot(labels["target"], 10)\n        loss = optax.softmax_cross_entropy(logits, target).mean()\n\n        # metrics\n        logs = dict(\n            acc=jnp.mean(jnp.argmax(logits, axis=-1) == labels["target"]),\n            loss=loss,\n        )\n\n        return loss, logs, self\n```\n\n**3.** Instantiate our `LinearClassifier` with an optimizer:\n\n```python\nmodel = LinearClassifier(\n    optimizer=optax.rmsprop(1e-3),\n)\n```\n\n**4.** Train the model using the `fit` method:\n\n```python\nmodel.fit(\n    inputs=X_train,\n    labels=y_train,\n    epochs=100,\n    steps_per_epoch=200,\n    batch_size=64,\n    validation_data=(X_test, y_test),\n    shuffle=True,\n    callbacks=[eg.callbacks.TensorBoard("summaries")]\n)\n```\n\n#### Using other JAX Frameworks\n\n<details>\n<summary>Show</summary>\n\nIt is straightforward to integrate other functional JAX libraries with this\nlow-level API, here is an example with Flax:\n\n```python\nimport elegy as eg\nimport flax.linen as nn\n\nclass LinearClassifier(eg.Model):\n    params: Mapping[str, Any] = eg.Parameter.node()\n    batch_stats: Mapping[str, Any] = eg.BatchStat.node()\n    next_key: eg.KeySeq\n\n    def __init__(self, module: nn.Module, **kwargs):\n        self.flax_module = module\n        super().__init__(**kwargs)\n\n    def init_step(self, key, inputs):\n        self.next_key = eg.KeySeq(key)\n\n        variables = self.flax_module.init(\n            {"params": self.next_key(), "dropout": self.next_key()}, x\n        )\n        self.params = variables["params"]\n        self.batch_stats = variables["batch_stats"]\n\n        self.optimizer = self.optimizer.init(self.parameters())\n\n    def test_step(self, inputs, labels):\n        # forward\n        variables = dict(\n            params=self.params,\n            batch_stats=self.batch_stats,\n        )\n        logits, variables = self.flax_module.apply(\n            variables,\n            inputs, \n            rngs={"dropout": self.next_key()}, \n            mutable=True,\n        )\n        self.batch_stats = variables["batch_stats"]\n        \n        # loss\n        target = jax.nn.one_hot(labels["target"], 10)\n        loss = optax.softmax_cross_entropy(logits, target).mean()\n\n        # logs\n        logs = dict(\n            accuracy=accuracy,\n            loss=loss,\n        )\n        return loss, logs, self\n```\n\n</details>\n\n### Examples\n\nCheck out the [/example](/examples) directory for some inspiration. To run an example, first install some requirements:\n\n```bash\npip install -r examples/requirements.txt\n```\n\nAnd the run it normally with python e.g.\n\n```bash\npython examples/flax/mnist_vae.py\n```\n\n## Contributing\n\nIf your are interested in helping improve Elegy check out the [Contributing Guide](https://poets-ai.github.io/elegy/guides/contributing).\n\n## Sponsors 💚\n* [Quansight](https://www.quansight.com) - paid development time\n\n## Citing Elegy\n\n\n**BibTeX**\n\n```\n@software{elegy2020repository,\n\ttitle        = {Elegy: A High Level API for Deep Learning in JAX},\n\tauthor       = {PoetsAI},\n\tyear         = 2021,\n\turl          = {https://github.com/poets-ai/elegy},\n\tversion      = {0.8.1}\n}\n```\n',
    'author': 'Cristian Garcia',
    'author_email': 'cgarcia.e88@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://poets-ai.github.io/elegy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)



# from .resnet import (
#     ResNet,
#     ResNet18,
#     ResNet34,
#     ResNet50,
#     ResNet101,
#     ResNet152,
#     ResNet200,
# )

# __all__ = [
#     "ResNet",
#     "ResNet18",
#     "ResNet34",
#     "ResNet50",
#     "ResNet101",
#     "ResNet152",
#     "ResNet200",
# ]



import typing as tp
from abc import ABC, abstractmethod

import jax
import jax.numpy as jnp
import treeo as to
import treex as tx

from elegy import types

A = tp.TypeVar("A")
M = tp.TypeVar("M", bound="CoreModule")
Me = tp.TypeVar("Me", bound="tx.Metric")


InitStep = tp.Callable[[M, jnp.ndarray, types.Inputs], M]
PredStep = tp.Callable[[M, types.Inputs], tp.Tuple[types.Outputs, M]]
TestStep = tp.Callable[[M, types.Inputs, types.Labels], tp.Tuple[types.Logs, M]]
TrainStep = tp.Callable[[M, types.Inputs, types.Labels], tp.Tuple[types.Logs, M]]


class CoreModuleMeta(to.TreeMeta):
    def construct(cls, obj: M, *args, **kwargs) -> M:
        obj = super().construct(obj, *args, **kwargs)

        if not hasattr(obj, "_called_init"):
            CoreModule.__init__(obj)

        return obj


class CoreModule(to.Tree, to.Immutable, to.Map, to.Copy, metaclass=CoreModuleMeta):

    initialized: bool
    distributed_strategy: "DistributedStrategy"

    def __init__(
        self,
        *,
        initialized: bool = False,
        distributed_strategy: tp.Optional["DistributedStrategy"] = None,
    ) -> None:
        from elegy.distributed_strategies import Eager

        self.initialized = initialized
        self.distributed_strategy = (
            distributed_strategy if distributed_strategy is not None else Eager()
        )
        self._called_init = None

    # ---------------------------------------------------------------------------
    # API
    # ---------------------------------------------------------------------------

    def reset_step(
        self: M,
    ) -> M:
        raise types.MissingMethod()

    def init_step(
        self: M,
        key: jnp.ndarray,
        inputs: tp.Any,
    ) -> M:
        raise types.MissingMethod()

    def predict_step(
        self: M,
        batch: tp.Any,
        batch_idx: int,
    ) -> tp.Tuple[types.Outputs, M]:
        raise types.MissingMethod()

    def test_step(
        self: M,
        batch: tp.Any,
        batch_idx: int,
    ) -> tp.Tuple[types.Logs, M]:
        raise types.MissingMethod()

    def train_step(
        self: M,
        batch: types.Inputs,
        batch_idx: int,
        epoch_idx: int,
    ) -> tp.Tuple[types.Logs, M]:
        raise types.MissingMethod()

    def tabulate(
        self,
        *args,
        summary_depth: int = 2,
        **kwargs,
    ) -> str:
        raise types.MissingMethod()

    # ---------------------------------------------------------------------------
    # Base methods
    # ---------------------------------------------------------------------------

    def mark_initialized(
        self: M,
    ) -> M:
        return self.replace(initialized=True)

    def set_distributed_strategy(
        self: M,
        distributed_strategy: "DistributedStrategy",
    ) -> M:
        return self.replace(distributed_strategy=distributed_strategy)

    # ---------------------------------------------------------------------------
    # Callback Methods
    # ---------------------------------------------------------------------------

    def on_epoch_begin(self: M, epoch: int, logs: tp.Optional[types.Logs] = None) -> M:
        """Called at the end of an epoch.

        Subclasses should override for any actions to run. This function should only
        be called during TRAIN mode.

        Arguments:
            epoch: integer, index of epoch.
            logs: dict, metric results for this training epoch, and for the
                validation epoch if validation is performed. Validation result keys
                are prefixed with `val_`.
        """
        return self

    def on_epoch_end(self: M, epoch: int, logs: tp.Optional[types.Logs] = None) -> M:
        """Called at the end of an epoch.

        Subclasses should override for any actions to run. This function should only
        be called during TRAIN mode.

        Arguments:
            epoch: integer, index of epoch.
            logs: dict, metric results for this training epoch, and for the
                validation epoch if validation is performed. Validation result keys
                are prefixed with `val_`.
        """
        return self

    def on_train_batch_begin(
        self: M, batch: int, logs: tp.Optional[types.Logs] = None
    ) -> M:
        """Called at the beginning of a training batch in `fit` methods.

        Subclasses should override for any actions to run.

        Arguments:
            batch: integer, index of batch within the current epoch.
            logs: dict. Has keys `batch` and `size` representing the current batch
                number and the size of the batch.
        """
        return self

    def on_train_batch_end(
        self: M, batch: int, logs: tp.Optional[types.Logs] = None
    ) -> M:
        """Called at the end of a training batch in `fit` methods.

        Subclasses should override for any actions to run.

        Arguments:
            batch: integer, index of batch within the current epoch.
            logs: dict. Metric results for this batch.
        """
        return self

    def on_test_batch_begin(
        self: M, batch: int, logs: tp.Optional[types.Logs] = None
    ) -> M:
        """Called at the beginning of a batch in `evaluate` methods.

        Also called at the beginning of a validation batch in the `fit`
        methods, if validation data is provided.

        Subclasses should override for any actions to run.

        Arguments:
            batch: integer, index of batch within the current epoch.
            logs: dict. Has keys `batch` and `size` representing the current batch
                number and the size of the batch.
        """
        return self

    def on_test_batch_end(
        self: M, batch: int, logs: tp.Optional[types.Logs] = None
    ) -> M:
        """Called at the end of a batch in `evaluate` methods.

        Also called at the end of a validation batch in the `fit`
        methods, if validation data is provided.

        Subclasses should override for any actions to run.

        Arguments:
            batch: integer, index of batch within the current epoch.
            logs: dict. Metric results for this batch.
        """
        return self

    def on_predict_batch_begin(
        self: M, batch: int, logs: tp.Optional[types.Logs] = None
    ) -> M:
        """Called at the beginning of a batch in `predict` methods.

        Subclasses should override for any actions to run.

        Arguments:
            batch: integer, index of batch within the current epoch.
            logs: dict. Has keys `batch` and `size` representing the current batch
                number and the size of the batch.
        """
        return self

    def on_predict_batch_end(
        self: M, batch: int, logs: tp.Optional[types.Logs] = None
    ) -> M:
        """Called at the end of a batch in `predict` methods.

        Subclasses should override for any actions to run.

        Arguments:
            batch: integer, index of batch within the current epoch.
            logs: dict. Metric results for this batch.
        """
        return self

    def on_train_begin(self: M, logs: tp.Optional[types.Logs] = None) -> M:
        """Called at the beginning of training.

        Subclasses should override for any actions to run.

        Arguments:
            logs: dict. Currently no data is passed to this argument for this method
                but that may change in the future.
        """
        return self

    def on_train_end(self: M, logs: tp.Optional[types.Logs] = None) -> M:
        """Called at the end of training.

        Subclasses should override for any actions to run.

        Arguments:
            logs: dict. Currently no data is passed to this argument for this method
                but that may change in the future.
        """
        return self

    def on_test_begin(self: M, logs: tp.Optional[types.Logs] = None) -> M:
        """Called at the beginning of evaluation or validation.

        Subclasses should override for any actions to run.

        Arguments:
            logs: dict. Currently no data is passed to this argument for this method
                but that may change in the future.
        """
        return self

    def on_test_end(self: M, logs: tp.Optional[types.Logs] = None) -> M:
        """Called at the end of evaluation or validation.

        Subclasses should override for any actions to run.

        Arguments:
            logs: dict. Currently no data is passed to this argument for this method
                but that may change in the future.
        """
        return self

    def on_predict_begin(self: M, logs: tp.Optional[types.Logs] = None) -> M:
        """Called at the beginning of prediction.

        Subclasses should override for any actions to run.

        Arguments:
            logs: dict. Currently no data is passed to this argument for this method
                but that may change in the future.
        """
        return self

    def on_predict_end(self: M, logs: tp.Optional[types.Logs] = None) -> M:
        """Called at the end of prediction.

        Subclasses should override for any actions to run.

        Arguments:
            logs: dict. Currently no data is passed to this argument for this method
                but that may change in the future.
        """
        return self


class DistributedStrategy(ABC):
    def from_local(self, module: M) -> M:
        return module

    def to_local(self, module: M) -> M:
        return module

    def lift_data(self, data: A) -> A:
        return data

    def lift_key(self, key: jnp.ndarray) -> jnp.ndarray:
        return key

    def lift_batch_size(self, batch_size: int) -> int:
        return batch_size

    def handle_post_init(self, module: M) -> M:
        return module

    def handle_metrics(
        self,
        metrics: Me,
    ) -> Me:
        return metrics

    def handle_mudule_and_grads(
        self,
        module: M,
        grads: tp.Any,
    ) -> tp.Tuple[M, tp.Any]:
        return module, grads

    @abstractmethod
    def init_step_fn(self, module: M) -> InitStep[M]:
        ...

    @abstractmethod
    def pred_step_fn(self, module: M) -> PredStep[M]:
        ...

    @abstractmethod
    def test_step_fn(self, module: M) -> TestStep[M]:
        ...

    @abstractmethod
    def train_step_fn(self, module: M) -> TrainStep[M]:
        ...

    # implement order methods, required so that DistributedStrategy can be
    # used as a key in a dict
    def __lt__(self, other):
        return self.__class__.__name__ < other.__class__.__name__

    def __le__(self, other):
        return self.__class__.__name__ <= other.__class__.__name__

    def __gt__(self, other):
        return self.__class__.__name__ > other.__class__.__name__

    def __ge__(self, other):
        return self.__class__.__name__ >= other.__class__.__name__
